import inspect
import logging
import pathlib
import sys
import traceback
import typing
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import sphinx.application
from sphinx import __version__ as __sphinx_version__
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE

from .. import types
from ..app import Sphinx
from ..config import SphinxConfig
from ..transforms import LineNumberTransform
from ..types import Uri
from ..util import send_error
from ..util import send_message

STATIC_DIR = (pathlib.Path(__file__).parent.parent / "static").resolve()
sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)

# Inject our own 'core' extensions into Sphinx
sphinx.application.builtin_extensions += (
    f"{__name__}.files",
    f"{__name__}.diagnostics",
    f"{__name__}.symbols",
    f"{__name__}.directives",
    f"{__name__}.roles",
    f"{__name__}.domains",
)


class SphinxHandler:
    """Responsible for implementing the JSON-RPC API exposed by the Sphinx agent."""

    def __init__(self):
        self.app: Optional[Sphinx] = None
        """The sphinx application instance"""

        self._content_overrides: Dict[Uri, str] = {}
        """Holds any additional content to inject into a build."""

        self._handlers: Dict[str, Tuple[Type, Callable]] = self._register_handlers()

    def get(self, method: str) -> Optional[Tuple[Type, Callable]]:
        """Return the handler for the given method - if possible.

        Parameters
        ----------
        method
           The name of the method

        Returns
        -------
        Optional[Tuple[Type, Callable]]
           A tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.

           If ``None``, the given method is unknown.

        """
        return self._handlers.get(method)

    def _register_handlers(self) -> Dict[str, Tuple[Type, Callable]]:
        """Return a map of all the handlers we provide.

        A handler

        - must be a method on this class
        - the must take a single parameter called ``request``
        - the type annotation for that parameter must correspond with a ``XXXRequest``
          class definition from the ``types`` module.

        Returns
        -------
        Dict[str, Tuple[Type, Callable]]
           A dictonary where the keys are the method names implemented by this class.
           Values are a tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.
        """
        handlers: Dict[str, Tuple[Type, Callable]] = {}

        for name in dir(self):
            method_func = getattr(self, name)
            if name.startswith("_") or not inspect.ismethod(method_func):
                continue

            parameters = inspect.signature(method_func).parameters
            if set(parameters.keys()) != {"request"}:
                continue

            request_type = typing.get_type_hints(method_func)["request"]
            if not all(
                [hasattr(request_type, "jsonrpc"), hasattr(request_type, "method")]
            ):
                continue

            handlers[request_type.method] = (request_type, method_func)

        return handlers

    def create_sphinx_app(self, request: types.CreateApplicationRequest):
        """Create a new sphinx application instance."""
        sphinx_config = SphinxConfig.fromcli(request.params.command)
        if sphinx_config is None:
            raise ValueError("Invalid build command")

        sphinx_args = sphinx_config.to_application_args()
        self.app = Sphinx(**sphinx_args)

        # Connect event handlers.
        self.app.connect("env-before-read-docs", self._cb_env_before_read_docs)
        self.app.connect("source-read", self._cb_source_read, priority=0)

        # TODO: Sphinx 7.x has introduced a `include-read` event
        # See: https://github.com/sphinx-doc/sphinx/pull/11657

        _enable_sync_scrolling(self.app)

        response = types.CreateApplicationResponse(
            id=request.id,
            result=types.SphinxInfo(
                version=__sphinx_version__,
                conf_dir=str(self.app.confdir),
                build_dir=str(self.app.outdir),
                builder_name=self.app.builder.name,
                src_dir=str(self.app.srcdir),
                dbpath=str(self.app.esbonio.db.path),
            ),
            jsonrpc=request.jsonrpc,
        )
        send_message(response)

    def _cb_env_before_read_docs(self, app: Sphinx, env, docnames: List[str]):
        """Used to add additional documents to the "to build" list."""

        is_building = set(docnames)

        for docname in env.found_docs - is_building:
            uri = Uri.for_file(env.doc2path(docname, base=True))
            if uri in self._content_overrides:
                docnames.append(docname)

    def _cb_source_read(self, app: Sphinx, docname: str, source):
        """Called whenever sphinx reads a file from disk."""

        uri = Uri.for_file(app.env.doc2path(docname, base=True))
        if (content := self._content_overrides.get(uri, None)) is not None:
            source[0] = content

    def build_sphinx_app(self, request: types.BuildRequest):
        """Trigger a Sphinx build."""

        if self.app is None:
            send_error(id=request.id, code=-32803, message="Sphinx app not initialized")
            return

        self._content_overrides = {
            Uri.parse(p): content
            for p, content in request.params.content_overrides.items()
        }

        try:
            self.app.build()

            response = types.BuildResponse(
                id=request.id,
                result=types.BuildResult(),
                jsonrpc=request.jsonrpc,
            )
            send_message(response)
        except Exception as exc:
            message = "".join(traceback.format_exception_only(type(exc), exc))
            sphinx_logger.error("sphinx-build failed", exc_info=True)
            send_error(
                id=request.id, code=-32603, message=f"sphinx-build failed: {message}"
            )

        finally:
            self.app._warncount = 0

    def notify_exit(self, request: types.ExitNotification):
        """Sent from the client to signal that the agent should exit."""
        sys.exit(0)


def _enable_sync_scrolling(app: Sphinx):
    """Given a Sphinx application, configure it so that we can support syncronised
    scrolling."""

    # Inline the JS code we need to enable sync scrolling.
    #
    # Yes this "bloats" every page in the generated docs, but is generally more robust
    # see: https://github.com/swyddfa/esbonio/issues/810
    webview_js = STATIC_DIR / "webview.js"
    app.add_js_file(None, body=webview_js.read_text())

    app.add_transform(LineNumberTransform)



from __future__ import annotations

import inspect
import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Coroutine
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Union

    from esbonio.server import Uri


class RoleProvider:
    """Base class for role providers."""

    def get_role(
        self, uri: Uri, name: str
    ) -> Union[Optional[types.Role], Coroutine[Any, Any, Optional[types.Role]]]:
        """Return the definition of the given role, if known.

        Parameters
        ----------
        uri
           The uri of the document in which the role name appears

        name
           The name of the role, as the user would type in a document
        """
        return None

    def suggest_roles(
        self, context: server.CompletionContext
    ) -> Union[
        Optional[List[types.Role]], Coroutine[Any, Any, Optional[List[types.Role]]]
    ]:
        """Givem a completion context, suggest roles that may be used."""
        return None


class RoleTargetProvider:
    """Base class for role target providers."""

    def suggest_targets(
        self, context: server.CompletionContext, **kwargs
    ) -> Union[
        Optional[List[lsp.CompletionItem]],
        Coroutine[Any, Any, Optional[List[lsp.CompletionItem]]],
    ]:
        """Givem a completion context, suggest role targets that may be used."""
        return None


class RolesFeature(server.LanguageFeature):
    """Backend support for roles.

    It's this language feature's responsibility to provide an API that exposes the
    information a frontend feature may want.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._role_providers: Dict[int, RoleProvider] = {}
        self._target_providers: Dict[str, RoleTargetProvider] = {}

    def add_role_provider(self, provider: RoleProvider):
        """Register a role provider.

        Parameters
        ----------
        provider
           The role provider
        """
        self._role_providers[id(provider)] = provider

    def add_target_provider(self, name: str, provider: RoleTargetProvider):
        """Register a role target provider.

        Parameters
        ----------
        provider
           The role target provider
        """
        if (existing := self._target_providers.get(name)) is not None:
            raise ValueError(
                f"RoleTargetProvider {provider!r} conflicts with existing "
                f"provider: {existing!r}"
            )

        self._target_providers[name] = provider

    async def suggest_roles(
        self, context: server.CompletionContext
    ) -> List[types.Role]:
        """Suggest roles that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context
        """
        items: List[types.Role] = []

        for provider in self._role_providers.values():
            try:
                result: Optional[List[types.Role]] = None

                aresult = provider.suggest_roles(context)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result:
                    items.extend(result)
            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.suggest_roles'", name, exc_info=True)

        return items

    async def get_role(self, uri: Uri, name: str) -> Optional[types.Role]:
        """Return the definition of the given role name.

        Parameters
        ----------
        uri
           The uri of the document in which the role name appears

        name
           The name of the role, as the user would type into a document.

        Returns
        -------
        Optional[types.Role]
           The role's definition, if known
        """
        for provider in self._role_providers.values():
            try:
                result: Optional[types.Role] = None

                aresult = provider.get_role(uri, name)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result is not None:
                    return result
            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.get_role'", name, exc_info=True)

        return None

    async def suggest_targets(
        self, context: server.CompletionContext, role_name: str
    ) -> List[lsp.CompletionItem]:
        """Suggest role targets that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context

        role_name
           The role to suggest targets for
        """
        if (role := await self.get_role(context.uri, role_name)) is None:
            self.logger.debug("Unknown role '%s'", role_name)
            return []

        targets = []
        self.logger.debug(
            "Suggesting targets for role: '%s' (%s)", role.name, role.implementation
        )

        for spec in role.target_providers:
            if (provider := self._target_providers.get(spec.name)) is None:
                self.logger.error("Unknown target provider: '%s'", spec.name)
                continue

            try:
                result: Optional[List[lsp.CompletionItem]] = None

                aresult = provider.suggest_targets(context, **spec.kwargs)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result is not None:
                    targets.extend(result)

            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.suggest_targets'", name, exc_info=True)

        return targets


def esbonio_setup(server: server.EsbonioLanguageServer):
    roles = RolesFeature(server)
    server.add_feature(roles)



import asyncio
import json
from typing import Dict
from typing import List
from typing import Optional

from lsprotocol import types

from esbonio.server import EsbonioLanguageServer
from esbonio.server import LanguageFeature
from esbonio.server import Uri
from esbonio.server.features.project_manager import ProjectManager


class SphinxSymbols(LanguageFeature):
    """Add support for ``textDocument/documentSymbol`` requests"""

    def __init__(self, server: EsbonioLanguageServer, manager: ProjectManager):
        super().__init__(server)
        self.manager = manager

    async def document_symbol(
        self, params: types.DocumentSymbolParams
    ) -> Optional[List[types.DocumentSymbol]]:
        """Called when a document symbols request is received."""

        uri = Uri.parse(params.text_document.uri)
        if (project := self.manager.get_project(uri)) is None:
            return None

        symbols = await project.get_document_symbols(uri)
        if len(symbols) == 0:
            return None

        root: List[types.DocumentSymbol] = []
        index: Dict[int, types.DocumentSymbol] = {}

        for id_, name, kind, detail, range_json, parent_id, _ in symbols:
            range_ = self.converter.structure(json.loads(range_json), types.Range)
            symbol = types.DocumentSymbol(
                name=name,
                kind=self.converter.structure(kind, types.SymbolKind),
                range=range_,
                selection_range=range_,
                detail=detail,
            )

            index[id_] = symbol

            if parent_id is None:
                root.append(symbol)
            else:
                parent = index[parent_id]
                if parent.children is None:
                    parent.children = [symbol]
                else:
                    parent.children.append(symbol)

        return root

    async def workspace_symbol(
        self, params: types.WorkspaceSymbolParams
    ) -> Optional[List[types.WorkspaceSymbol]]:
        """Called when a workspace symbol request is received."""

        tasks = []
        for project in self.manager.projects.values():
            tasks.append(
                asyncio.create_task(project.get_workspace_symbols(params.query))
            )

        symbols = await asyncio.gather(*tasks)
        result: List[types.WorkspaceSymbol] = []

        for batch in symbols:
            for uri_str, name, kind, detail, range_json, container in batch:
                uri = Uri.parse(uri_str)
                range_ = self.converter.structure(json.loads(range_json), types.Range)

                if detail not in {"", name}:
                    display_name = f"{name} {detail}"
                else:
                    display_name = name

                result.append(
                    types.WorkspaceSymbol(
                        location=types.Location(uri=str(uri), range=range_),
                        name=display_name,
                        kind=self.converter.structure(kind, types.SymbolKind),
                        container_name=container,
                    )
                )

        return result


def esbonio_setup(server: EsbonioLanguageServer, project_manager: ProjectManager):
    symbols = SphinxSymbols(server, project_manager)
    server.add_feature(symbols)



from __future__ import annotations

import importlib
import inspect
import pathlib
import typing
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Set
from typing import Type

from lsprotocol import types

from . import Uri

if typing.TYPE_CHECKING:
    from .server import EsbonioLanguageServer


def create_language_server(
    server_cls: Type[EsbonioLanguageServer], modules: Iterable[str], *args, **kwargs
) -> EsbonioLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    server_cls:
       The class definition to create the server from.
    modules:
       The list of modules that should be loaded.
    args, kwargs:
       Any additional arguments that should be passed to the language server's
       constructor.
    """
    server = server_cls(*args, **kwargs)

    for module in modules:
        _load_module(server, module)

    _configure_lsp_methods(server)
    _configure_completion(server)

    return server


def _configure_lsp_methods(server: EsbonioLanguageServer):
    """Configure method handlers for the portions of the LSP spec we support."""

    @server.feature(types.INITIALIZE)
    async def on_initialize(ls: EsbonioLanguageServer, params: types.InitializeParams):
        ls.initialize(params)
        await call_features(ls, "initialize", params)

    @server.feature(types.INITIALIZED)
    async def on_initialized(
        ls: EsbonioLanguageServer, params: types.InitializedParams
    ):
        await ls.initialized(params)
        await call_features(ls, "initialized", params)

    @server.feature(types.SHUTDOWN)
    async def on_shutdown(ls: EsbonioLanguageServer, params: None):
        ls.lsp_shutdown(params)
        await call_features(ls, "shutdown", params)

    @server.feature(types.TEXT_DOCUMENT_DID_CHANGE)
    async def on_document_change(
        ls: EsbonioLanguageServer, params: types.DidChangeTextDocumentParams
    ):
        await call_features(ls, "document_change", params)

    @server.feature(types.TEXT_DOCUMENT_DID_CLOSE)
    async def on_document_close(
        ls: EsbonioLanguageServer, params: types.DidCloseTextDocumentParams
    ):
        await call_features(ls, "document_close", params)

    @server.feature(types.TEXT_DOCUMENT_DID_OPEN)
    async def on_document_open(
        ls: EsbonioLanguageServer, params: types.DidOpenTextDocumentParams
    ):
        await call_features(ls, "document_open", params)

    @server.feature(types.TEXT_DOCUMENT_DID_SAVE)
    async def on_document_save(
        ls: EsbonioLanguageServer, params: types.DidSaveTextDocumentParams
    ):
        # Record the version number of the document
        doc = ls.workspace.get_document(params.text_document.uri)
        doc.saved_version = doc.version or 0

        await call_features(ls, "document_save", params)

    @server.feature(
        types.TEXT_DOCUMENT_DIAGNOSTIC,
        types.DiagnosticOptions(
            identifier="esbonio",
            inter_file_dependencies=True,
            workspace_diagnostics=True,
        ),
    )
    async def on_document_diagnostic(
        ls: EsbonioLanguageServer, params: types.DocumentDiagnosticParams
    ):
        """Handle a ``textDocument/diagnostic`` request."""
        doc_uri = Uri.parse(params.text_document.uri).resolve()
        items = []

        for (_, uri), diags in ls._diagnostics.items():
            if uri.resolve() == doc_uri:
                items.extend(diags)

        # TODO: Detect no changes and send 'unchanged' responses
        return types.RelatedFullDocumentDiagnosticReport(
            items=items,
            kind=types.DocumentDiagnosticReportKind.Full,
        )

    @server.feature(types.WORKSPACE_DIAGNOSTIC)
    async def on_workspace_diagnostic(
        ls: EsbonioLanguageServer, params: types.WorkspaceDiagnosticParams
    ):
        """Handle a ``workspace/diagnostic`` request."""
        diagnostics: Dict[Uri, List[types.Diagnostic]] = {}

        for (_, uri), diags in ls._diagnostics.items():
            diagnostics.setdefault(uri, []).extend(diags)

        # TODO: Detect no changes and send 'unchanged' responses
        reports = []
        for uri, items in diagnostics.items():
            reports.append(
                types.WorkspaceFullDocumentDiagnosticReport(
                    uri=str(uri),
                    items=items,
                    kind=types.DocumentDiagnosticReportKind.Full,
                )
            )

        # Typing issues should be fixed in a future version of lsprotocol
        # see: https://github.com/microsoft/lsprotocol/pull/285
        return types.WorkspaceDiagnosticReport(items=reports)  # type: ignore[arg-type]

    @server.feature(types.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
    async def on_document_symbol(
        ls: EsbonioLanguageServer, params: types.DocumentSymbolParams
    ):
        result = await return_first_result(ls, "document_symbol", params)
        return result

    @server.feature(types.WORKSPACE_SYMBOL)
    async def on_workspace_symbol(
        ls: EsbonioLanguageServer, params: types.WorkspaceSymbolParams
    ):
        result = await return_first_result(ls, "workspace_symbol", params)
        if len(result) == 0:
            return None

        return result

    @server.feature(types.WORKSPACE_DID_CHANGE_CONFIGURATION)
    async def on_did_change_configuration(
        ls: EsbonioLanguageServer, params: types.DidChangeConfigurationParams
    ):
        ls.logger.debug("%s: %s", types.WORKSPACE_DID_CHANGE_CONFIGURATION, params)
        await ls.configuration.update_workspace_configuration()

    @server.feature(types.WORKSPACE_DID_CHANGE_WATCHED_FILES)
    async def on_did_change_watched_files(
        ls: EsbonioLanguageServer, params: types.DidChangeWatchedFilesParams
    ):
        ls.logger.debug("%s: %s", types.WORKSPACE_DID_CHANGE_WATCHED_FILES, params)
        # TODO: Handle deleted files.
        paths = [pathlib.Path(Uri.parse(event.uri)) for event in params.changes]
        await ls.configuration.update_file_configuration(paths)


def _configure_completion(server: EsbonioLanguageServer):
    """Configuration completion handlers."""

    trigger_characters: Set[str] = set()

    for _, feature in server:
        if feature.completion_trigger is None:
            continue

        trigger_characters.update(feature.completion_trigger.characters)

    @server.feature(
        types.TEXT_DOCUMENT_COMPLETION,
        types.CompletionOptions(
            trigger_characters=list(trigger_characters),
            resolve_provider=True,
        ),
    )
    async def on_completion(ls: EsbonioLanguageServer, params: types.CompletionParams):
        uri = params.text_document.uri
        pos = params.position
        doc = ls.workspace.get_text_document(uri)
        language = ls.get_language_at(doc, pos)

        items = []

        for cls, feature in ls:
            if not feature.completion_trigger:
                continue

            context = feature.completion_trigger(
                uri=Uri.parse(uri),
                params=params,
                document=doc,
                language=language,
                client_capabilities=ls.client_capabilities,
            )

            if context is None:
                continue

            ls.logger.debug("%s", context)
            name = f"{cls.__name__}"

            try:
                result = feature.completion(context)
                if inspect.isawaitable(result):
                    result = await result
            except Exception:
                ls.logger.exception("Error in '%s.complete' handler", name)
                continue

            for item in result or []:
                item.data = {"source_feature": name, **(item.data or {})}  # type: ignore
                items.append(item)

        if len(items) > 0:
            return types.CompletionList(is_incomplete=False, items=items)

    @server.feature(types.COMPLETION_ITEM_RESOLVE)
    def on_completion_resolve(
        ls: EsbonioLanguageServer, item: types.CompletionItem
    ) -> types.CompletionItem:
        # source = (item.data or {}).get("source_feature", "")  # type: ignore
        # feature = ls.get_feature(source)

        # if not feature:
        #     ls.logger.error(
        #         "Unable to resolve completion item, unknown source: '%s'", source
        #     )
        #     return item

        # return feature.completion_resolve(item)
        return item


async def call_features(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                await result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


async def gather_results(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features, gathering all results into a list."""
    results: List[Any] = []
    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                items = await result
            else:
                items = result

            if isinstance(items, list):
                results.extend(items)
            elif items is not None:
                results.append(items)

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)

    return results


async def return_first_result(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features, returning the first non ``None`` result we find."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                result = await result

            if result is not None:
                return result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


def _load_module(server: EsbonioLanguageServer, modname: str):
    """Load an extension module by calling its ``esbonio_setup`` function, if it exists."""

    try:
        module = importlib.import_module(modname)
    except ImportError:
        server.logger.error("Unable to import module '%s'", modname, exc_info=True)
        return None

    setup = getattr(module, "esbonio_setup", None)
    if setup is None:
        server.logger.debug(
            "Skipping module '%s', missing 'esbonio_setup' function", modname
        )
        return None

    server.load_extension(modname, setup)



from __future__ import annotations

import typing

from lsprotocol import types

from esbonio import server
from esbonio.server.features import directives
from esbonio.server.features.project_manager import ProjectManager

if typing.TYPE_CHECKING:
    from typing import List
    from typing import Optional


class SphinxDirectives(directives.DirectiveProvider):
    """Support for directives in a sphinx project."""

    def __init__(self, manager: ProjectManager):
        self.manager = manager

    async def suggest_directives(
        self, context: server.CompletionContext
    ) -> Optional[List[directives.Directive]]:
        """Given a completion context, suggest directives that may be used."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        # Does the document have a default domain set?
        results = await project.find_symbols(
            uri=str(context.uri.resolve()),
            kind=types.SymbolKind.Class.value,
            detail="default-domain",
        )
        if len(results) > 0:
            default_domain = results[0][1]
        else:
            default_domain = None

        primary_domain = await project.get_config_value("primary_domain")
        active_domain = default_domain or primary_domain or "py"

        result: List[directives.Directive] = []
        for name, implementation in await project.get_directives():
            # std: directives can be used unqualified
            if name.startswith("std:"):
                short_name = name.replace("std:", "")
                result.append(
                    directives.Directive(name=short_name, implementation=implementation)
                )

            # Also suggest unqualified versions of directives from the currently active domain.
            elif name.startswith(f"{active_domain}:"):
                short_name = name.replace(f"{active_domain}:", "")
                result.append(
                    directives.Directive(name=short_name, implementation=implementation)
                )

            result.append(
                directives.Directive(name=name, implementation=implementation)
            )

        return result


def esbonio_setup(
    project_manager: ProjectManager,
    directive_feature: directives.DirectiveFeature,
):
    provider = SphinxDirectives(project_manager)
    directive_feature.add_provider(provider)



import pathlib

import pytest

from esbonio.server import Uri

TEST_DIR = pathlib.Path(__file__).parent


@pytest.fixture(scope="session")
def uri_for():
    """Helper function for returning the uri for a given file in the ``tests/``
    directory."""

    def fn(*args):
        path = (TEST_DIR / pathlib.Path(*args)).resolve()
        assert path.exists(), f"{path} does not exist"
        return Uri.for_file(str(path))

    return fn



from __future__ import annotations

import asyncio
import traceback
import typing
import uuid
from functools import partial

import attrs
import lsprotocol.types as lsp

from esbonio import server
from esbonio.server import Uri
from esbonio.sphinx_agent import types

from .client import ClientState
from .config import SphinxConfig

if typing.TYPE_CHECKING:
    from typing import Callable
    from typing import Dict
    from typing import Optional

    from esbonio.server.features.project_manager import ProjectManager

    from .client import SphinxClient

    SphinxClientFactory = Callable[["SphinxManager", "SphinxConfig"], "SphinxClient"]


@attrs.define
class ClientCreatedNotification:
    """The payload of a ``sphinx/clientCreated`` notification"""

    id: str
    """The client's id"""

    scope: str
    """The scope at which the client was created."""

    config: SphinxConfig
    """The final configuration."""


@attrs.define
class AppCreatedNotification:
    """The payload of a ``sphinx/appCreated`` notification"""

    id: str
    """The client's id"""

    application: types.SphinxInfo
    """Details about the created application."""


@attrs.define
class ClientErroredNotification:
    """The payload of a ``sphinx/clientErrored`` notification"""

    id: str
    """The client's id"""

    error: str
    """Short description of the error."""

    detail: str
    """Detailed description of the error."""


@attrs.define
class ClientDestroyedNotification:
    """The payload of ``sphinx/clientDestroyed`` notification."""

    id: str
    """The client's id"""


class SphinxManager(server.LanguageFeature):
    """Responsible for managing Sphinx application instances."""

    def __init__(
        self,
        client_factory: SphinxClientFactory,
        project_manager: ProjectManager,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.client_factory = client_factory
        """Used to create new Sphinx client instances."""

        self.project_manager = project_manager
        """The project manager instance to use."""

        self.clients: Dict[str, Optional[SphinxClient]] = {
            # Prevent any clients from being created in the global scope.
            "": None,
        }
        """Holds currently active Sphinx clients."""

        self._events = server.EventSource(self.logger)
        """The SphinxManager can emit events."""

        self._pending_builds: Dict[str, asyncio.Task] = {}
        """Holds tasks that will trigger a build after a given delay if not cancelled."""

        self._progress_tokens: Dict[str, str] = {}
        """Holds work done progress tokens."""

    def add_listener(self, event: str, handler):
        self._events.add_listener(event, handler)

    async def document_change(self, params: lsp.DidChangeTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        self._pending_builds[client.id] = asyncio.create_task(
            self.trigger_build_after(uri, client.id, delay=2)
        )

    async def document_open(self, params: lsp.DidOpenTextDocumentParams):
        # Ensure that a Sphinx app instance is created the first time a document in a
        # given project is opened.
        if (uri := Uri.parse(params.text_document.uri)) is not None:
            await self.get_client(uri)

    async def document_save(self, params: lsp.DidSaveTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        await self.trigger_build(uri)

    async def shutdown(self, params: None):
        """Called when the server is instructed to ``shutdown``."""

        # Stop any existing clients.
        tasks = []
        for client in self.clients.values():
            if client:
                self.logger.debug("Stopping SphinxClient: %s", client)
                tasks.append(asyncio.create_task(client.stop()))

        await asyncio.gather(*tasks)

    async def trigger_build_after(self, uri: Uri, app_id: str, delay: float):
        """Trigger a build for the given uri after the given delay."""
        await asyncio.sleep(delay)

        self._pending_builds.pop(app_id)
        self.logger.debug("Triggering build")
        await self.trigger_build(uri)

    async def trigger_build(self, uri: Uri):
        """Trigger a build for the relevant Sphinx application for the given uri."""

        client = await self.get_client(uri)
        if client is None or client.state != ClientState.Running:
            return

        if (project := self.project_manager.get_project(uri)) is None:
            return

        # Pass through any unsaved content to the Sphinx agent.
        content_overrides: Dict[str, str] = {}
        known_src_uris = await project.get_src_uris()

        for src_uri in known_src_uris:
            doc = self.server.workspace.get_document(str(src_uri))
            doc_version = doc.version or 0
            saved_version = getattr(doc, "saved_version", 0)

            if saved_version < doc_version:
                content_overrides[str(src_uri)] = doc.source

        await self.start_progress(client)

        try:
            result = await client.build(content_overrides=content_overrides)
        except Exception as exc:
            self.server.show_message(f"{exc}", lsp.MessageType.Error)
            return
        finally:
            self.stop_progress(client)

        # Notify listeners.
        self._events.trigger("build", client, result)

    async def get_client(self, uri: Uri) -> Optional[SphinxClient]:
        """Given a uri, return the relevant sphinx client instance for it."""

        scope = self.server.configuration.scope_for(uri)
        if scope not in self.clients:
            self.logger.debug("No client found, creating new subscription")
            self.server.configuration.subscribe(
                "esbonio.sphinx",
                SphinxConfig,
                partial(self._create_or_replace_client, uri),
                scope=uri,
            )
            # The first few callers in a given scope will miss out, but that shouldn't matter
            # too much
            return None

        if (client := self.clients[scope]) is None:
            self.logger.debug("No applicable client for uri: %s", uri)
            return None

        return await client

    async def _create_or_replace_client(
        self, uri: Uri, event: server.ConfigChangeEvent[SphinxConfig]
    ):
        """Create or replace thesphinx client instance for the given config.

        Parameters
        ----------
        uri
           The uri for which the sphinx client was originally created for

        event
           The configuration change event
        """

        config = event.value

        # Do not try and create clients in the global scope
        if event.scope == "":
            return

        # If there was a previous client, stop it.
        if (previous_client := self.clients.pop(event.scope, None)) is not None:
            self.server.lsp.notify(
                "sphinx/clientDestroyed",
                ClientDestroyedNotification(id=previous_client.id),
            )
            self.server.run_task(previous_client.stop())

        resolved = config.resolve(uri, self.server.workspace, self.logger)
        if resolved is None:
            self.clients[event.scope] = None
            return

        self.clients[event.scope] = client = self.client_factory(self, resolved)
        client.add_listener("state-change", partial(self._on_state_change, event.scope))

        self.server.lsp.notify(
            "sphinx/clientCreated",
            ClientCreatedNotification(id=client.id, scope=event.scope, config=resolved),
        )
        self.logger.debug("Client created for scope %s", event.scope)

        # Start the client
        await client

    def _on_state_change(
        self,
        scope: str,
        client: SphinxClient,
        old_state: ClientState,
        new_state: ClientState,
    ):
        """React to state changes in the client."""

        if old_state == ClientState.Starting and new_state == ClientState.Running:
            if (sphinx_info := client.sphinx_info) is not None:
                self.project_manager.register_project(scope, client.db)
                self.server.lsp.notify(
                    "sphinx/appCreated",
                    AppCreatedNotification(id=client.id, application=sphinx_info),
                )

        if new_state == ClientState.Errored:
            error = ""
            detail = ""

            if (exc := getattr(client, "exception", None)) is not None:
                error = f"{type(exc).__name__}: {exc}"
                detail = "".join(
                    traceback.format_exception(type(exc), exc, exc.__traceback__)
                )

            self.server.lsp.show_message(error, lsp.MessageType.Error)
            self.server.lsp.notify(
                "sphinx/clientErrored",
                ClientErroredNotification(id=client.id, error=error, detail=detail),
            )

    async def start_progress(self, client: SphinxClient):
        """Start reporting work done progress for the given client."""

        token = str(uuid.uuid4())
        self.logger.debug("Starting progress: '%s'", token)

        try:
            await self.server.progress.create_async(token)
        except Exception as exc:
            self.logger.debug("Unable to create progress token: %s", exc)
            return

        self._progress_tokens[client.id] = token
        self.server.progress.begin(
            token,
            lsp.WorkDoneProgressBegin(title="sphinx-build", cancellable=False),
        )

    def stop_progress(self, client: SphinxClient):
        if (token := self._progress_tokens.pop(client.id, None)) is None:
            return

        self.server.progress.end(token, lsp.WorkDoneProgressEnd(message="Finished"))

    def report_progress(self, client: SphinxClient, progress: types.ProgressParams):
        """Report progress done for the given client."""

        if client.state not in {ClientState.Running, ClientState.Building}:
            return

        if (token := self._progress_tokens.get(client.id, None)) is None:
            return

        self.server.progress.report(
            token,
            lsp.WorkDoneProgressReport(
                message=progress.message,
                percentage=progress.percentage,
                cancellable=False,
            ),
        )



from __future__ import annotations

import pathlib
import sqlite3
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Literal
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union


class Database:
    @dataclass
    class Column:
        name: str
        dtype: str
        notnull: bool = field(default=False)
        default: Optional[Any] = field(default=None)
        pk: int = field(default=0)

        @property
        def definition(self):
            # TODO: Is there a way to do this via a prepared statement?
            return f"{self.name} {self.dtype}"

    @dataclass
    class Table:
        name: str
        columns: List[Database.Column]

        @property
        def create_statement(self):
            """Return the SQL statement required to create this table."""
            # TODO: Is there a way to do this via a prepared statement?
            columns = ",".join([c.definition for c in self.columns])
            return "".join([f"CREATE TABLE {self.name} (", columns, ");"])

    def __init__(self, dbpath: Union[pathlib.Path, Literal[":memory:"]]):
        self.path = dbpath

        if isinstance(self.path, pathlib.Path) and not self.path.parent.exists():
            self.path.parent.mkdir(parents=True)

        self.db = sqlite3.connect(self.path)

        # Ensure that Write Ahead Logging is enabled.
        self.db.execute("PRAGMA journal_mode(WAL)")

        self._checked_tables: Set[str] = set()

    def _get_table(self, name: str) -> Optional[Table]:
        """Get the table with the given name, if it exists."""
        # TODO: SQLite does not seem to like '?' syntax in this statement...
        cursor = self.db.execute(f"PRAGMA table_info({name});")
        rows = cursor.fetchall()

        if len(rows) == 0:
            # Table does not exist
            return None

        columns = [
            self.Column(name=name, dtype=type_, notnull=notnull, default=default, pk=pk)
            for (_, name, type_, notnull, default, pk) in rows
        ]

        return self.Table(name=name, columns=columns)

    def _create_table(self, table: Table):
        """Create the given table."""
        cursor = self.db.cursor()

        # TODO: Is there a way to do this via a prepared statement?
        cursor.execute(f"DROP TABLE IF EXISTS {table.name}")
        cursor.execute(table.create_statement)
        self.db.commit()

    def clear_table(self, table: Table, **kwargs):
        """Clear the given table

        Parameters
        ----------
        kwargs
           Constraints to limit the rows that get cleared
        """

        # TODO: Is there a way to pass the table name as a '?' parameter?
        base_query = f"DELETE FROM {table.name}"  # noqa: S608
        where: List[str] = []
        parameters: List[Any] = []

        for param, value in kwargs.items():
            where.append(f"{param} = ?")
            parameters.append(value)

        if where:
            conditions = " AND ".join(where)
            query = " ".join([base_query, "WHERE", conditions])
        else:
            query = base_query

        cursor = self.db.cursor()
        cursor.execute(query, tuple(parameters))
        self.db.commit()

    def ensure_table(self, table: Table):
        """Ensure that the given table exists in the database.

        If the table *does* exist, but has the wrong shape, it will be dropped and
        recreated.
        """
        # If we've already checked the table, then there's nothing to do
        if table.name in self._checked_tables:
            return

        if (existing := self._get_table(table.name)) is None:
            self._create_table(table)
            return

        # Are the tables compatible?
        if len(existing.columns) != len(table.columns):
            self._create_table(table)
        else:
            for existing_col, col in zip(existing.columns, table.columns):
                if existing_col.name != col.name or existing_col.dtype != col.dtype:
                    self._create_table(table)
                    break

        self._checked_tables.add(table.name)

    def insert_values(self, table: Table, values: List[Tuple]):
        """Insert the given values into the given table."""

        if len(values) == 0:
            return

        cursor = self.db.cursor()

        placeholder = "(" + ",".join(["?" for _ in range(len(values[0]))]) + ")"
        cursor.executemany(f"INSERT INTO {table.name} VALUES {placeholder}", values)  # noqa: S608
        self.db.commit()



from __future__ import annotations

import typing

from lsprotocol import types

from esbonio import server
from esbonio.server.features.directives import DirectiveFeature
from esbonio.server.features.directives import completion
from esbonio.sphinx_agent.types import RST_DIRECTIVE

if typing.TYPE_CHECKING:
    from typing import List
    from typing import Optional


class RstDirectives(server.LanguageFeature):
    """A frontend to directives for reStructuredText syntax."""

    def __init__(self, directives: DirectiveFeature, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.directives = directives
        self._insert_behavior = "replace"

    completion_trigger = server.CompletionTrigger(
        patterns=[RST_DIRECTIVE],
        languages={"rst"},
        characters={".", "`"},
    )

    def initialized(self, params: types.InitializedParams):
        """Called once the initial handshake between client and server has finished."""
        self.configuration.subscribe(
            "esbonio.server.completion",
            server.CompletionConfig,
            self.update_configuration,
        )

    def update_configuration(
        self, event: server.ConfigChangeEvent[server.CompletionConfig]
    ):
        """Called when the user's configuration is updated."""
        self._insert_behavior = event.value.preferred_insert_behavior

    async def completion(
        self, context: server.CompletionContext
    ) -> Optional[List[types.CompletionItem]]:
        """Provide completion suggestions for directives."""

        groups = context.match.groupdict()

        # Are we completing a directive's options?
        if "directive" not in groups:
            return await self.complete_options(context)

        # Don't offer completions for targets
        if (groups["name"] or "").startswith("_"):
            return None

        # Are we completing the directive's argument?
        directive_end = context.match.span()[0] + len(groups["directive"])
        complete_directive = groups["directive"].endswith("::")

        if complete_directive and directive_end < context.position.character:
            return await self.complete_arguments(context)

        return await self.complete_directives(context)

    async def complete_options(self, context: server.CompletionContext):
        return None

    async def complete_arguments(self, context: server.CompletionContext):
        return None

    async def complete_directives(
        self, context: server.CompletionContext
    ) -> Optional[List[types.CompletionItem]]:
        """Return completion suggestions for the available directives."""

        render_func = completion.get_directive_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        for directive in await self.directives.suggest_directives(context):
            if (item := render_func(context, directive)) is not None:
                items.append(item)

        if len(items) > 0:
            return items

        return None


def esbonio_setup(esbonio: server.EsbonioLanguageServer, directives: DirectiveFeature):
    rst_directives = RstDirectives(directives, esbonio)
    esbonio.add_feature(rst_directives)



from __future__ import annotations

import pathlib
import typing

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

if typing.TYPE_CHECKING:
    from typing import Optional
    from typing import Set


EXPECTED = {
    "function",
    "module",
    "option",
    "program",
    "image",
    "toctree",
    "c:macro",
    "c:function",
    "py:function",
    "py:module",
    "std:program",
    "std:option",
}

RST_EXPECTED = EXPECTED.copy()
MYST_EXPECTED = {"eval-rst", *EXPECTED}

UNEXPECTED = {
    "macro",
    "restructuredtext-test-directive",
}

RST_UNEXPECTED = {"eval-rst", *UNEXPECTED}
MYST_UNEXPECTED = UNEXPECTED.copy()


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        (".", None, None),
        ("..", RST_EXPECTED, RST_UNEXPECTED),
        (".. ", RST_EXPECTED, RST_UNEXPECTED),
        (".. d", RST_EXPECTED, RST_UNEXPECTED),
        (".. code-b", RST_EXPECTED, RST_UNEXPECTED),
        (".. codex-block:: ", None, None),
        (".. c:", RST_EXPECTED, RST_UNEXPECTED),
        (".. _some_label:", None, None),
        ("   .", None, None),
        ("   ..", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. ", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. d", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. doctest:: ", None, None),
        ("   .. code-b", RST_EXPECTED, RST_UNEXPECTED),
        ("   .. codex-block:: ", None, None),
        ("   .. _some_label:", None, None),
        ("   .. c:", RST_EXPECTED, RST_UNEXPECTED),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_rst_directive_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer directive completions in rst
    documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "directives.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your note here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("`", None, None),
        ("``", None, None),
        ("```", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{d", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{code-b", MYST_EXPECTED, MYST_UNEXPECTED),
        ("```{codex-block} ", None, None),
        ("```{c:", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   `", None, None),
        ("   ``", None, None),
        ("   ```", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{d", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{doctest}", None, None),
        ("   ```{code-b", MYST_EXPECTED, MYST_UNEXPECTED),
        ("   ```{codex-block}", None, None),
        ("   ```{c:", MYST_EXPECTED, MYST_UNEXPECTED),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_myst_directive_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "directives.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your note here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected



from __future__ import annotations

import typing

from lsprotocol import types

from esbonio import server
from esbonio.server.features.roles import RolesFeature
from esbonio.server.features.roles import completion
from esbonio.sphinx_agent.types import RST_DIRECTIVE
from esbonio.sphinx_agent.types import RST_ROLE

if typing.TYPE_CHECKING:
    from typing import List
    from typing import Optional


class RstRoles(server.LanguageFeature):
    """A frontend to roles for reStructuredText syntax."""

    def __init__(self, roles: RolesFeature, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.roles = roles
        self._insert_behavior = "replace"

    completion_trigger = server.CompletionTrigger(
        patterns=[RST_ROLE],
        languages={"rst"},
        characters={":", "`", "<", "/"},
    )

    def initialized(self, params: types.InitializedParams):
        """Called once the initial handshake between client and server has finished."""
        self.configuration.subscribe(
            "esbonio.server.completion",
            server.CompletionConfig,
            self.update_configuration,
        )

    def update_configuration(
        self, event: server.ConfigChangeEvent[server.CompletionConfig]
    ):
        """Called when the user's configuration is updated."""
        self._insert_behavior = event.value.preferred_insert_behavior

    async def completion(
        self, context: server.CompletionContext
    ) -> Optional[List[types.CompletionItem]]:
        """Provide completion suggestions for roles."""

        groups = context.match.groupdict()
        target = groups["target"]

        # All text matched by the regex
        text = context.match.group(0)
        start, end = context.match.span()

        if target:
            target_index = start + text.find(target)

            # Only trigger target completions if the request was made from within
            # the target part of the role.
            if target_index <= context.position.character <= end:
                return await self.complete_targets(context)

        # If there's no indent, then this can only be a
        # role definition
        indent = context.match.group(1)
        if indent == "":
            return await self.complete_roles(context)

        # Otherwise, search backwards until we find a blank line or an unindent
        # so that we can determine the appropriate context.
        linum = context.position.line - 1

        try:
            line = context.doc.lines[linum]
        except IndexError:
            return await self.complete_roles(context)

        while linum >= 0 and line.startswith(indent):
            linum -= 1
            line = context.doc.lines[linum]

        # Unless we are within a directive's options block, we should offer role
        # suggestions
        if RST_DIRECTIVE.match(line):
            return []

        return await self.complete_roles(context)

    async def complete_targets(
        self, context: server.CompletionContext
    ) -> Optional[List[types.CompletionItem]]:
        """Provide completion suggestions for role targets."""

        render_func = completion.get_role_target_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        role_name = context.match.group("name")
        for target in await self.roles.suggest_targets(context, role_name):
            if (item := render_func(context, target)) is not None:
                items.append(item)

        return items if len(items) > 0 else None

    async def complete_roles(
        self, context: server.CompletionContext
    ) -> Optional[List[types.CompletionItem]]:
        """Return completion suggestions for the available roles"""

        render_func = completion.get_role_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        for role in await self.roles.suggest_roles(context):
            if (item := render_func(context, role)) is not None:
                items.append(item)

        if len(items) > 0:
            return items

        return None


def esbonio_setup(esbonio: server.EsbonioLanguageServer, roles: RolesFeature):
    rst_roles = RstRoles(roles, esbonio)
    esbonio.add_feature(rst_roles)



from __future__ import annotations

import logging
import pathlib
import typing

from sphinx.application import Sphinx as _Sphinx
from sphinx.util import console
from sphinx.util import logging as sphinx_logging_module
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE

from . import types
from .database import Database
from .log import DiagnosticFilter

if typing.TYPE_CHECKING:
    from typing import IO
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Set
    from typing import Tuple
    from typing import Type

    from sphinx.domains import Domain

    RoleDefinition = Tuple[str, Any, List[types.Role.TargetProvider]]

sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)
sphinx_log_setup = sphinx_logging_module.setup


def setup_logging(app: Sphinx, status: IO, warning: IO):
    # Run the usual setup
    sphinx_log_setup(app, status, warning)

    # Attach our diagnostic filter to the warning handler.
    for handler in sphinx_logger.handlers:
        if handler.level == logging.WARNING:
            handler.addFilter(app.esbonio.log)


class Esbonio:
    """Esbonio specific functionality."""

    db: Database

    log: DiagnosticFilter

    def __init__(self, dbpath: pathlib.Path, app: _Sphinx):
        self.db = Database(dbpath)
        self.log = DiagnosticFilter(app)

        self._roles: List[RoleDefinition] = []
        """Roles captured during Sphinx startup."""

    def add_role(
        self,
        name: str,
        role: Any,
        target_providers: Optional[List[types.Role.TargetProvider]] = None,
    ):
        """Register a role with esbonio.

        Parameters
        ----------
        name
           The name of the role, as the user would type in a document

        role
           The role's implementation

        target_providers
           A list of target providers for the role
        """
        self._roles.append((name, role, target_providers or []))

    @staticmethod
    def create_role_target_provider(name: str, **kwargs) -> types.Role.TargetProvider:
        """Create a new role target provider

        Parameters
        ----------
        name
           The name of the provider

        kwargs
           Additional arguments to pass to the provider instance

        Returns
        -------
        types.Role.TargetProvider
           The target provider
        """
        return types.Role.TargetProvider(name, kwargs)


class Sphinx(_Sphinx):
    """An extended sphinx application that integrates with esbonio."""

    esbonio: Esbonio

    def __init__(self, *args, **kwargs):
        # Disable color codes
        console.nocolor()

        # Add in esbonio specific functionality
        self.esbonio = Esbonio(
            dbpath=pathlib.Path(kwargs["outdir"], "esbonio.db").resolve(),
            app=self,
        )

        # Override sphinx's usual logging setup function
        sphinx_logging_module.setup = setup_logging  # type: ignore

        super().__init__(*args, **kwargs)

    def add_role(self, name: str, role: Any, override: bool = False):
        super().add_role(name, role, override)
        self.esbonio.add_role(name, role)

    def add_domain(self, domain: Type[Domain], override: bool = False) -> None:
        super().add_domain(domain, override)

        target_types: Dict[str, Set[str]] = {}

        for obj_name, item_type in domain.object_types.items():
            for role_name in item_type.roles:
                target_type = f"{domain.name}:{obj_name}"
                target_types.setdefault(role_name, set()).add(target_type)

        for name, role in domain.roles.items():
            providers = []
            if (item_types := target_types.get(name)) is not None:
                providers.append(
                    self.esbonio.create_role_target_provider(
                        "objects", obj_types=list(item_types)
                    )
                )

            self.esbonio.add_role(f"{domain.name}:{name}", role, providers)



import logging
import os
import pathlib
from typing import Optional

import pytest
from lsprotocol.types import WorkspaceFolder
from pygls.workspace import Workspace

from esbonio.server import Uri
from esbonio.server.features.sphinx_manager import SphinxConfig

logger = logging.getLogger(__name__)


# Default values
CWD = os.path.join(".", "path", "to", "workspace")[1:]
PYTHON_CMD = ["/bin/python"]
BUILD_CMD = ["sphinx-build", "-M", "html", "src", "dest"]
PYPATH = [pathlib.Path("/path/to/site-packages/esbonio")]


def mk_uri(path: str) -> str:
    return str(Uri.for_file(path))


@pytest.mark.parametrize(
    "uri, workspace, config, expected",
    [
        (  # If everything is specified, resolve should be a no-op
            "file:///path/to/workspace/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                cwd=CWD,
                python_path=PYPATH,
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                cwd=CWD,
                python_path=PYPATH,
            ),
        ),
        (  # If no cwd is given, and there is no available workspace root the config
            # should be considered invalid
            "file:///path/to/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            None,
        ),
        (  # If the workspace is empty, we should still be able to progress as long as
            # the user provides a cwd
            "file:///path/to/file.rst",
            Workspace(None),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
                cwd=CWD,
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
                cwd=CWD,
            ),
        ),
        (  # If only a ``root_uri`` is given use that.
            "file:///path/to/workspace/file.rst",
            Workspace(mk_uri(CWD)),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
                cwd=CWD,
            ),
        ),
        (  # Assuming that the requested uri resides within it.
            "file:///path/to/other/workspace/file.rst",
            Workspace(mk_uri(CWD)),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            None,
        ),
        (  # Otherwise, prefer workspace_folders.
            "file:///path/to/workspace/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[WorkspaceFolder(mk_uri(CWD), "workspace")],
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
                cwd=CWD,
            ),
        ),
        (  # Handle multi-root scenarios.
            "file:///path/to/workspace-b/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[
                    WorkspaceFolder("file:///path/to/workspace-a", "workspace-a"),
                    WorkspaceFolder("file:///path/to/workspace-b", "workspace-b"),
                ],
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
                cwd=os.path.join(".", "path", "to", "workspace-b")[1:],
            ),
        ),
        (  # Again, make sure the requested uri resides within the workspace.
            "file:///path/for/workspace-c/file.rst",
            Workspace(
                "file:///path/to",
                workspace_folders=[
                    WorkspaceFolder("file:///path/to/workspace-a", "workspace-a"),
                    WorkspaceFolder("file:///path/to/workspace-b", "workspace-b"),
                ],
            ),
            SphinxConfig(
                python_command=PYTHON_CMD,
                build_command=BUILD_CMD,
                python_path=PYPATH,
            ),
            None,
        ),
    ],
)
def test_resolve(
    uri: str,
    workspace: Workspace,
    config: SphinxConfig,
    expected: Optional[SphinxConfig],
):
    """Ensure that we can resolve a user's configuration correctly.

    Parameters
    ----------
    uri
       The uri the config should be resolved relative to

    workspace
       The workspace in which to resolve the configuration

    config
       The base configuration to resolve

    expected
       The expected outcome
    """
    assert config.resolve(Uri.parse(uri), workspace, logger) == expected



"""Type definitions for the sphinx agent.

This is the *only* file shared between the agent itself and the parent language server.
For this reason this file *cannot* import anything from Sphinx.
"""

import dataclasses
import os
import pathlib
import re
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib import parse

from .lsp import Diagnostic
from .lsp import DiagnosticSeverity
from .lsp import Location
from .lsp import Position
from .lsp import Range
from .roles import MYST_ROLE
from .roles import RST_DEFAULT_ROLE
from .roles import RST_ROLE
from .roles import Role

__all__ = (
    "Diagnostic",
    "DiagnosticSeverity",
    "Location",
    "MYST_ROLE",
    "Position",
    "RST_DEFAULT_ROLE",
    "RST_ROLE",
    "Range",
    "Role",
)

MYST_DIRECTIVE: "re.Pattern" = re.compile(
    r"""
    (\s*)                             # directives can be indented
    (?P<directive>
      ```(`*)?                        # directives start with at least 3 ` chars
      (?!\w)                          # -- regular code blocks are not directives
      [{]?                            # followed by an opening brace
      (?P<name>[^}]+)?                # directives have a name
      [}]?                            # directives are closed with a closing brace
    )
    (\s+(?P<argument>.*?)\s*$)?       # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete MyST directives.

This does **not** include any options or content that may be included with the
initial declaration.
"""


RST_DIRECTIVE: "re.Pattern" = re.compile(
    r"""
    (\s*)                             # directives can be indented
    (?P<directive>
      \.\.                            # directives start with a comment
      [ ]?                            # followed by a space
      (?P<substitution>\|             # this could be a substitution definition
        (?P<substitution_text>[^|]+)?
      \|?)?
      [ ]?
      (?P<name>([\w-]|:(?!:))+)?      # directives have a name
      (::)?                           # directives end with '::'
    )
    ([\s]+(?P<argument>.*?)\s*$)?     # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete directives.

This does **not** include any options or content that may be included underneath
the initial declaration. A number of named capture groups are available.

``name``
   The name of the directive, not including the domain prefix.

``directive``
   Everything that makes up a directive, from the initial ``..`` up to and including the
   ``::`` characters.

``argument``
   All argument text.

``substitution``
   If the directive is part of a substitution definition, this group will contain
"""


RST_DIRECTIVE_OPTION: "re.Pattern" = re.compile(
    r"""
    (?P<indent>\s+)       # directive options must be indented
    (?P<option>
      :                   # options start with a ':'
      (?P<name>[\w-]+)?   # options have a name
      :?                  # options end with a ':'
    )
    (\s*
      (?P<value>.*)       # options can have a value
    )?
    """,
    re.VERBOSE,
)
"""A regular expression used to detect and parse partial and complete directive options.

A number of named capture groups are available

``name``
   The name of the option

``option``
   The name of the option including the surrounding ``:`` characters.

``indent``
   The whitespace characters making preceeding the initial ``:`` character

``value``
   The value passed to the option

"""


IS_WIN = os.name == "nt"
SCHEME = re.compile(r"^[a-zA-Z][a-zA-Z\d+.-]*$")
RE_DRIVE_LETTER_PATH = re.compile(r"^(\/?)([a-zA-Z]:)")


# TODO: Look into upstreaming this into pygls
#       - if it works out
#       - when pygls drops 3.7 (Uri uses the := operator)
@dataclasses.dataclass(frozen=True)
class Uri:
    """Helper class for working with URIs."""

    scheme: str

    authority: str

    path: str

    query: str

    fragment: str

    def __post_init__(self):
        """Basic validation."""
        if self.scheme is None:
            raise ValueError("URIs must have a scheme")

        if not SCHEME.match(self.scheme):
            raise ValueError("Invalid scheme")

        if self.authority and self.path and (not self.path.startswith("/")):
            raise ValueError("Paths with an authority must start with a slash '/'")

        if self.path and self.path.startswith("//") and (not self.authority):
            raise ValueError(
                "Paths without an authority cannot start with two slashes '//'"
            )

    def __eq__(self, other):
        if type(other) is not type(self):
            return False

        if self.scheme != other.scheme:
            return False

        if self.authority != other.authority:
            return False

        if self.query != other.query:
            return False

        if self.fragment != other.fragment:
            return False

        if IS_WIN and self.scheme == "file":
            # Filepaths on windows are case in-sensitive
            if self.path.lower() != other.path.lower():
                return False

        elif self.path != other.path:
            return False

        return True

    def __hash__(self):
        if IS_WIN and self.scheme == "file":
            # Filepaths on windows are case in-sensitive
            path = self.path.lower()
        else:
            path = self.path

        return hash((self.scheme, self.authority, path, self.query, self.fragment))

    def __fspath__(self):
        """Return the file system representation of this uri.

        This makes Uri instances compatible with any function that expects an
        ``os.PathLike`` object!
        """
        # TODO: Should we raise an exception if scheme != "file"?
        return self.as_fs_path(preserve_case=True)

    def __str__(self):
        return self.as_string()

    def __truediv__(self, other):
        return self.join(other)

    @classmethod
    def create(
        cls,
        *,
        scheme: str = "",
        authority: str = "",
        path: str = "",
        query: str = "",
        fragment: str = "",
    ) -> "Uri":
        """Create a uri with the given attributes."""

        if scheme in {"http", "https", "file"}:
            if not path.startswith("/"):
                path = f"/{path}"

        return cls(
            scheme=scheme,
            authority=authority,
            path=path,
            query=query,
            fragment=fragment,
        )

    @classmethod
    def parse(cls, uri: str) -> "Uri":
        """Parse the given uri from its string representation."""
        scheme, authority, path, _, query, fragment = parse.urlparse(uri)
        return cls.create(
            scheme=parse.unquote(scheme),
            authority=parse.unquote(authority),
            path=parse.unquote(path),
            query=parse.unquote(query),
            fragment=parse.unquote(fragment),
        )

    def resolve(self) -> "Uri":
        """Return the fully resolved version of this Uri."""

        # This operation only makes sense for file uris
        if self.scheme != "file":
            return Uri.parse(str(self))

        return Uri.for_file(pathlib.Path(self).resolve())

    @classmethod
    def for_file(cls, filepath: Union[str, "os.PathLike[str]"]) -> "Uri":
        """Create a uri based on the given filepath."""

        fpath = os.fspath(filepath)
        if IS_WIN:
            fpath = fpath.replace("\\", "/")

        if fpath.startswith("//"):
            authority, *path = fpath[2:].split("/")
            fpath = "/".join(path)
        else:
            authority = ""

        return cls.create(scheme="file", authority=authority, path=fpath)

    @property
    def fs_path(self) -> Optional[str]:
        """Return the equivalent fs path."""
        return self.as_fs_path()

    def where(self, **kwargs) -> "Uri":
        """Return an transformed version of this uri where certain components of the uri
        have been replace with the given arguments.

        Passing a value of ``None`` will remove the given component entirely.
        """
        keys = {"scheme", "authority", "path", "query", "fragment"}
        valid_keys = keys.copy() & kwargs.keys()

        current = {k: getattr(self, k) for k in keys}
        replacements = {k: kwargs[k] for k in valid_keys}

        return Uri.create(**{**current, **replacements})

    def join(self, path: str) -> "Uri":
        """Join this Uri's path component with the given path and return the resulting
        uri.

        Parameters
        ----------
        path
           The path segment to join

        Returns
        -------
        Uri
           The resulting uri
        """

        if not self.path:
            raise ValueError("This uri has no path")

        if IS_WIN:
            fs_path = self.fs_path
            if fs_path is None:
                raise ValueError("Unable to join paths, fs_path is None")

            joined = os.path.normpath(os.path.join(fs_path, path))
            new_path = self.for_file(joined).path
        else:
            new_path = os.path.normpath(os.path.join(self.path, path))

        return self.where(path=new_path)

    def as_fs_path(self, preserve_case: bool = False) -> Optional[str]:
        """Return the file system path correspondin with this uri."""
        if self.path:
            path = _normalize_path(self.path, preserve_case)

            if self.authority and len(path) > 1:
                path = f"//{self.authority}{path}"

            # Remove the leading `/` from windows paths
            elif RE_DRIVE_LETTER_PATH.match(path):
                path = path[1:]

            if IS_WIN:
                path = path.replace("/", "\\")

            return path

        return None

    def as_string(self, encode=True) -> str:
        """Return a string representation of this Uri.

        Parameters
        ----------
        encode
           If ``True`` (the default), encode any special characters.

        Returns
        -------
        str
           The string representation of the Uri
        """

        # See: https://github.com/python/mypy/issues/10740
        encoder: Callable[[str], str] = parse.quote if encode else _replace_chars  # type: ignore[assignment]

        if authority := self.authority:
            usercred, *auth = authority.split("@")
            if len(auth) > 0:
                *user, cred = usercred.split(":")
                if len(user) > 0:
                    usercred = encoder(":".join(user)) + f":{encoder(cred)}"
                else:
                    usercred = encoder(usercred)
                authority = "@".join(auth)
            else:
                usercred = ""

            authority = authority.lower()
            *auth, port = authority.split(":")
            if len(auth) > 0:
                authority = encoder(":".join(auth)) + f":{port}"
            else:
                authority = encoder(authority)

            if usercred:
                authority = f"{usercred}@{authority}"

        scheme_separator = ""
        if authority or self.scheme == "file":
            scheme_separator = "//"

        if path := self.path:
            path = encoder(_normalize_path(path))

        if query := self.query:
            query = encoder(query)

        if fragment := self.fragment:
            fragment = encoder(fragment)

        parts = [
            f"{self.scheme}:",
            scheme_separator,
            authority if authority else "",
            path if path else "",
            f"?{query}" if query else "",
            f"#{fragment}" if fragment else "",
        ]
        return "".join(parts)


def _replace_chars(segment: str) -> str:
    """Replace a certain subset of characters in a uri segment"""
    return segment.replace("#", "%23").replace("?", "%3F")


def _normalize_path(path: str, preserve_case: bool = False) -> str:
    """Normalise the path segment of a Uri.

    Parameters
    ----------
    path
       The path to normalise.

    preserve_case
       If ``True``, preserve the case of the drive label on Windows.
       If ``False``, the drive label will be lowercased.

    Returns
    -------
    str
       The normalised path.
    """

    # normalize to fwd-slashes on windows,
    # on other systems bwd-slashes are valid
    # filename character, eg /f\oo/ba\r.txt
    if IS_WIN:
        path = path.replace("\\", "/")

    # Normalize drive paths to lower case
    if (not preserve_case) and (match := RE_DRIVE_LETTER_PATH.match(path)):
        path = match.group(1) + match.group(2).lower() + path[match.end() :]

    return path


# -- DB Types
#
# These represent the structure of data as stored in the SQLite database
Directive = Tuple[str, Optional[str], Optional[str]]
Symbol = Tuple[  # Represents either a document symbol or workspace symbol depending on context.
    int,  # id
    str,  # name
    int,  # kind
    str,  # detail
    str,  # range - as json object
    Optional[int],  # parent_id
    int,  # order_id
]


@dataclasses.dataclass
class CreateApplicationParams:
    """Parameters of a ``sphinx/createApp`` request."""

    command: List[str]
    """The ``sphinx-build`` command to base the app instance on."""


@dataclasses.dataclass
class CreateApplicationRequest:
    """A ``sphinx/createApp`` request."""

    id: Union[int, str]

    params: CreateApplicationParams

    method: str = "sphinx/createApp"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class SphinxInfo:
    """Represents information about an instance of the Sphinx application."""

    version: str
    """The version of Sphinx being used."""

    conf_dir: str
    """The folder containing the project's conf.py"""

    build_dir: str
    """The folder containing the Sphinx application's build output"""

    builder_name: str
    """The name of the builder in use"""

    src_dir: str
    """The folder containing the source files for the project"""

    dbpath: str
    """The filepath the database is stored in."""


@dataclasses.dataclass
class CreateApplicationResponse:
    """A ``sphinx/createApp`` response."""

    id: Union[int, str]

    result: SphinxInfo

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildParams:
    """Parameters of a ``sphinx/build`` request."""

    filenames: List[str] = dataclasses.field(default_factory=list)

    force_all: bool = False

    content_overrides: Dict[str, str] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class BuildResult:
    """Results from a ``sphinx/build`` request."""

    diagnostics: Dict[str, List[Diagnostic]] = dataclasses.field(default_factory=dict)
    """Any diagnostics associated with the project."""


@dataclasses.dataclass
class BuildRequest:
    """A ``sphinx/build`` request."""

    id: Union[int, str]

    params: BuildParams

    method: str = "sphinx/build"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildResponse:
    """A ``sphinx/build`` response."""

    id: Union[int, str]

    result: BuildResult

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class LogMessageParams:
    """Parameters of a ``window/logMessage`` notification."""

    type: int

    message: str


@dataclasses.dataclass
class LogMessage:
    """A ``window/logMessage`` notification"""

    params: LogMessageParams

    method: str = "window/logMessage"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class ProgressParams:
    message: Optional[str] = None

    percentage: Optional[int] = None


@dataclasses.dataclass
class ProgressMessage:
    """A ``$/progress`` notification"""

    params: ProgressParams

    method: str = "$/progress"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class ExitNotification:
    """An ``exit`` notification"""

    params: None

    method: str = "exit"

    jsonrpc: str = dataclasses.field(default="2.0")


METHOD_TO_MESSAGE_TYPE = {
    BuildRequest.method: BuildRequest,
    ExitNotification.method: ExitNotification,
    CreateApplicationRequest.method: CreateApplicationRequest,
}
METHOD_TO_RESPONSE_TYPE = {
    BuildRequest.method: BuildResponse,
    ExitNotification.method: None,
    CreateApplicationRequest.method: CreateApplicationResponse,
}



from __future__ import annotations

import json
import pathlib
import typing

import aiosqlite

from esbonio.server import Uri
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Tuple
    from typing import Type
    from typing import TypeVar
    from typing import Union

    import cattrs

    T = TypeVar("T")


class Project:
    """Represents a documentation project."""

    def __init__(self, dbpath: Union[str, pathlib.Path], converter: cattrs.Converter):
        self.converter = converter
        self.dbpath = dbpath
        self._connection: Optional[aiosqlite.Connection] = None

    async def close(self):
        if self._connection is not None:
            await self._connection.close()

    async def get_db(self) -> aiosqlite.Connection:
        if self._connection is None:
            self._connection = await aiosqlite.connect(self.dbpath)

        return self._connection

    def load_as(self, o: str, t: Type[T]) -> T:
        return self.converter.structure(json.loads(o), t)

    async def get_src_uris(self) -> List[Uri]:
        """Return all known source uris."""
        db = await self.get_db()

        query = "SELECT uri FROM files"
        async with db.execute(query) as cursor:
            results = await cursor.fetchall()
            return [Uri.parse(s[0]) for s in results]

    async def get_build_path(self, src_uri: Uri) -> Optional[str]:
        """Get the build path associated with the given ``src_uri``."""
        db = await self.get_db()

        query = "SELECT urlpath FROM files WHERE uri = ?"
        async with db.execute(query, (str(src_uri.resolve()),)) as cursor:
            if (result := await cursor.fetchone()) is None:
                return None

            return result[0]

    async def get_config_value(self, name: str) -> Optional[Any]:
        """Return the requested configuration value, if available."""

        db = await self.get_db()
        query = "SELECT value FROM config WHERE name = ?"
        cursor = await db.execute(query, (name,))

        if (row := await cursor.fetchone()) is None:
            return None

        (value,) = row
        return json.loads(value)

    async def get_directives(self) -> List[Tuple[str, Optional[str]]]:
        """Get the directives known to Sphinx."""
        db = await self.get_db()

        query = "SELECT name, implementation FROM directives"
        cursor = await db.execute(query)
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_role(self, name: str) -> Optional[types.Role]:
        """Get the roles known to Sphinx."""
        db = await self.get_db()

        query = "SELECT * FROM roles WHERE name = ?"
        cursor = await db.execute(query, (name,))
        result = await cursor.fetchone()

        return types.Role.from_db(self.load_as, *result) if result is not None else None

    async def get_roles(self) -> List[Tuple[str, Optional[str]]]:
        """Get the roles known to Sphinx."""
        db = await self.get_db()

        query = "SELECT name, implementation FROM roles"
        cursor = await db.execute(query)
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_document_symbols(self, src_uri: Uri) -> List[types.Symbol]:
        """Get the symbols for the given file."""
        db = await self.get_db()
        query = (
            "SELECT id, name, kind, detail, range, parent_id, order_id "
            "FROM symbols WHERE uri = ?"
        )
        cursor = await db.execute(query, (str(src_uri.resolve()),))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def find_symbols(self, **kwargs) -> List[types.Symbol]:
        """Find symbols which match the given criteria."""
        db = await self.get_db()
        base_query = (
            "SELECT id, name, kind, detail, range, parent_id, order_id FROM symbols"
        )
        where: List[str] = []
        parameters: List[Any] = []

        for param, value in kwargs.items():
            where.append(f"{param} = ?")
            parameters.append(value)

        if where:
            conditions = " AND ".join(where)
            query = " ".join([base_query, "WHERE", conditions])
        else:
            query = base_query

        cursor = await db.execute(query, tuple(parameters))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_workspace_symbols(
        self, query: str
    ) -> List[Tuple[str, str, int, str, str, str]]:
        """Return all the workspace symbols matching the given query string"""

        db = await self.get_db()
        sql_query = """\
SELECT
    child.uri,
    child.name,
    child.kind,
    child.detail,
    child.range,
    COALESCE(parent.name, '') AS container_name
FROM
    symbols child
LEFT JOIN
    symbols parent ON (child.parent_id = parent.id AND child.uri = parent.uri)
WHERE
    child.name like ? or child.detail like ?;"""

        query_str = f"%{query}%"
        cursor = await db.execute(sql_query, (query_str, query_str))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_diagnostics(self) -> Dict[Uri, List[Dict[str, Any]]]:
        """Get diagnostics for the project."""
        db = await self.get_db()
        cursor = await db.execute("SELECT * FROM diagnostics")
        results: Dict[Uri, List[Dict[str, Any]]] = {}

        for uri_str, item in await cursor.fetchall():
            uri = Uri.parse(uri_str)
            diagnostic = json.loads(item)
            results.setdefault(uri, []).append(diagnostic)

        return results



from __future__ import annotations

import typing

from sphinx import addnodes

from .. import types
from ..app import Database
from ..app import Sphinx
from ..util import as_json

if typing.TYPE_CHECKING:
    from typing import Dict
    from typing import Optional
    from typing import Tuple

OBJECTS_TABLE = Database.Table(
    "objects",
    [
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="display", dtype="TEXT"),
        Database.Column(name="domain", dtype="TEXT"),
        Database.Column(name="objtype", dtype="TEXT"),
        Database.Column(name="docname", dtype="TEXT"),
        Database.Column(name="description", dtype="TEXT"),
        Database.Column(name="location", dtype="JSON"),
    ],
)


class DomainObjects:
    """Discovers and indexes domain objects."""

    def __init__(self, app: Sphinx):
        self._info: Dict[
            Tuple[str, str, str, str], Tuple[Optional[str], Optional[str]]
        ] = {}

        app.connect("builder-inited", self.init_db)
        app.connect("object-description-transform", self.object_defined)
        app.connect("build-finished", self.commit)

    def init_db(self, app: Sphinx):
        """Prepare the database."""
        app.esbonio.db.ensure_table(OBJECTS_TABLE)

    def commit(self, app, exc):
        """Commit changes to the database.

        The only way to guarantee we discover all objects, from all domains correctly,
        is to call the ``get_objects()`` method on each domain. This means we process
        every object, every time we build.

        I will be *very* surprised if this never becomes a performance issue, but we
        will have to think of a smarter approach when it comes to it.
        """
        app.esbonio.db.clear_table(OBJECTS_TABLE)
        rows = []

        for name, domain in app.env.domains.items():
            for objname, dispname, objtype, docname, _, _ in domain.get_objects():
                desc, location = self._info.get(
                    (objname, name, objtype, docname), (None, None)
                )
                rows.append(
                    (objname, str(dispname), name, objtype, docname, desc, location)
                )

        app.esbonio.db.insert_values(OBJECTS_TABLE, rows)
        self._info.clear()

    def object_defined(
        self, app: Sphinx, domain: str, objtype: str, content: addnodes.desc_content
    ):
        """Record additional information about a domain object.

        Despite having a certain amount of structure to them (thanks to the API),
        domains can still do arbitrary things - take a peek at the implementations of
        the ``std``, ``py`` and ``cpp`` domains!

        So while this will never be perfect, this method is called each time the
        ``object-description-transform`` event is fired and attempts to extract the
        object's description and precise location.

        The trick however, is linking these items up with the correct object
        """

        sig = content.parent[0]

        try:
            name = sig["ids"][0]  # type: ignore[index]
        except Exception:
            return

        docname = app.env.docname
        description = content.astext()

        if (source := sig.source) is not None and (line := sig.line) is not None:
            location = as_json(
                types.Location(
                    uri=str(types.Uri.for_file(source)),
                    range=types.Range(
                        start=types.Position(line=line, character=0),
                        end=types.Position(line=line + 1, character=0),
                    ),
                )
            )
        else:
            location = None

        key = (name, domain, objtype, docname)
        self._info[key] = (description, location)


def setup(app: Sphinx):
    DomainObjects(app)



from esbonio.sphinx_agent.types import Uri

from ._configuration import ConfigChangeEvent
from .events import EventSource
from .feature import CompletionConfig
from .feature import CompletionContext
from .feature import CompletionTrigger
from .feature import LanguageFeature
from .server import EsbonioLanguageServer
from .server import EsbonioWorkspace
from .server import __version__
from .setup import create_language_server

__all__ = (
    "__version__",
    "ConfigChangeEvent",
    "CompletionConfig",
    "CompletionContext",
    "CompletionTrigger",
    "EsbonioLanguageServer",
    "EsbonioWorkspace",
    "EventSource",
    "LanguageFeature",
    "Uri",
    "create_language_server",
)



from __future__ import annotations

import asyncio
import inspect
import json
import pathlib
import traceback
import typing
from functools import partial
from typing import Generic
from typing import TypeVar

import attrs
from lsprotocol import types
from pygls.capabilities import get_capability

from . import Uri

T = TypeVar("T")

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Awaitable
    from typing import Callable
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Set
    from typing import Type
    from typing import Union

    from .server import EsbonioLanguageServer

    ConfigurationCallback = Callable[
        ["ConfigChangeEvent"], Union[Awaitable[None], None]
    ]


try:
    import tomllib as toml
except ImportError:
    import tomli as toml  # type: ignore[no-redef]


@attrs.define(frozen=True)
class Subscription(Generic[T]):
    """Represents a configuration subscription"""

    section: str
    """The configuration section."""

    spec: Type[T]
    """The subscription's class definition."""

    callback: ConfigurationCallback
    """The subscription's callback."""

    workspace_scope: str
    """The corresponding workspace scope for the subscription."""

    file_scope: str
    """The corresponding file scope for the subscription."""


@attrs.define
class ConfigChangeEvent(Generic[T]):
    """Is sent to subscribers when a configuration change occurs."""

    scope: str
    """The scope at which this configuration change occured."""

    value: T
    """The latest configuration value."""

    previous: Optional[T] = None
    """The previous configuration value, (if any)."""


class Configuration:
    """Manages the configuration values for the server.

    This will looks configuration in the following locations, in descending order of
    priority.

    - The server's ``initializationOptions``
    - ``workspace/configuration`` requests
    - A relevant configuration file

    **Note:** This assumes that a user's client is *either* going to be using
    ``initializationOptions`` *or* ``workspace/configuration`` not both. Any wierdness
    caused by combining both will (probably) not be fixed.

    **Note:** ``initializationOptions`` will implictly work at the global scope. There
    will be not multi-root/multi-project support using ``initializationOptions``. Use
    ``workspace/configuration`` or config files instead.
    """

    def __init__(self, server: EsbonioLanguageServer):
        self.server = server
        """The language server"""

        self.logger = server.logger.getChild("Configuration")
        """The logger instance to use"""

        self._initialization_options: Dict[str, Any] = {}
        """The received initializaion options (if any)"""

        self._workspace_config: Dict[str, Dict[str, Any]] = {}
        """The cached workspace configuration."""

        self._file_config: Dict[str, Dict[str, Any]] = {}
        """The cached configuration coming from configuration files."""

        self._subscriptions: Dict[Subscription, Any] = {}
        """Subscriptions and their last known value"""

        self._tasks: Set[asyncio.Task] = set()
        """Holds tasks that are currently executing an async config handler."""

    @property
    def initialization_options(self):
        return self._initialization_options

    @initialization_options.setter
    def initialization_options(self, value):
        """Ensure the init options are namespaced."""
        if value is None or value == {}:
            return

        if "esbonio" not in value:
            value = dict(esbonio=value)

        self._initialization_options = value

    @property
    def converter(self):
        return self.server.converter

    @property
    def workspace(self):
        return self.server.workspace

    @property
    def supports_workspace_config(self):
        """Indicates if the client supports ``workspace/configuration`` requests."""
        return get_capability(
            self.server.client_capabilities, "workspace.configuration", False
        )

    def subscribe(
        self,
        section: str,
        spec: Type[T],
        callback: ConfigurationCallback,
        scope: Optional[Uri] = None,
    ):
        """Subscribe to updates to the given configuration section.

        Parameters
        ----------
        section
           The configuration section to subscribe to

        spec
           A class representing the configuration values of interest

        callback
           The function to call when changes are detected.

        scope
           An optional uri, specifying the scope at which to lookup the configuration.
        """
        file_scope = self._uri_to_file_scope(scope)
        workspace_scope = self._uri_to_workspace_scope(scope)
        subscription = Subscription(
            section, spec, callback, workspace_scope, file_scope
        )

        if subscription in self._subscriptions:
            self.logger.debug("Ignoring duplicate subscription: %s", subscription)
            return

        self._subscriptions[subscription] = None

        # Once the server is ready, update all the subscriptions
        self.server.ready.add_done_callback(self._notify_subscriptions)

    def _notify_subscriptions(self, *args):
        """Notify subscriptions about configuration changes, if necessary."""

        for subscription, previous_value in self._subscriptions.items():
            value = self._get_config(
                subscription.section,
                subscription.spec,
                subscription.workspace_scope,
                subscription.file_scope,
            )

            # No need to notify if nothing has changed
            self.logger.debug("Previous: %s", previous_value)
            self.logger.debug("Current: %s", value)
            if previous_value == value:
                continue

            self._subscriptions[subscription] = value
            change_event = ConfigChangeEvent(
                scope=max(
                    [subscription.file_scope, subscription.workspace_scope], key=len
                ),
                value=value,
                previous=previous_value,
            )
            self.logger.info("%s", change_event)

            try:
                ret = subscription.callback(change_event)
                if inspect.iscoroutine(ret):
                    task = asyncio.create_task(ret)
                    task.add_done_callback(partial(self._finish_task, subscription))

            except Exception:
                self.logger.error(
                    "Error in configuration callback: %s",
                    subscription.callback,
                    exc_info=True,
                )

    def _finish_task(self, subscription: Subscription, task: asyncio.Task[None]):
        """Cleanup a finished task."""
        self._tasks.discard(task)

        if (exc := task.exception()) is not None:
            self.logger.error(
                "Error in async configuration handler '%s'\n%s",
                subscription.callback,
                "".join(traceback.format_exception(type(exc), exc, exc.__traceback__)),
            )

    def get(self, section: str, spec: Type[T], scope: Optional[Uri] = None) -> T:
        """Get the requested configuration section.

        Parameters
        ----------
        section
           The configuration section to retrieve

        spec
           A class representing the configuration values of interest

        scope
           An optional uri, specifying the scope in which to lookup the configuration.

        Returns
        -------
        T
           The requested configuration section parsed as an instance of ``T``.
        """
        file_scope = self._uri_to_file_scope(scope)
        workspace_scope = self._uri_to_workspace_scope(scope)

        return self._get_config(section, spec, workspace_scope, file_scope)

    def scope_for(self, uri: Uri) -> str:
        """Return the configuration scope that corresponds to the given uri.

        Parameters
        ----------
        uri
           The uri to return the scope for

        Returns
        -------
        str
           The scope corresponding with the given uri
        """

        file_scope = self._uri_to_file_scope(uri)
        workspace_scope = self._uri_to_workspace_scope(uri)

        return max([file_scope, workspace_scope], key=len)

    def _get_config(
        self, section: str, spec: Type[T], workspace_scope: str, file_scope: str
    ) -> T:
        """Get the requested configuration section."""

        self.logger.debug("File scope: '%s'", file_scope)
        self.logger.debug("Workspace scope: '%s'", workspace_scope)

        # To keep things simple, this method assumes that all available config is already
        # cached locally. Populating the cache is handled elsewhere.
        file_config = self._file_config.get(file_scope, {})
        workspace_config = self._workspace_config.get(workspace_scope, {})

        # Combine and resolve all the config sources - order matters!
        config = _merge_configs(
            file_config, workspace_config, self._initialization_options
        )
        # self.logger.debug("Full config: %s", json.dumps(config, indent=2))

        # Extract the requested section.
        config_section = config
        for name in section.split("."):
            config_section = config_section.get(name, {})

        self.logger.debug("Resolved config: %s", json.dumps(config_section, indent=2))

        try:
            value = self.converter.structure(config_section, spec)
            self.logger.debug("%s", value)

            return value
        except Exception:
            self.logger.error(
                "Unable to parse configuation as '%s', using defaults",
                spec.__name__,
                exc_info=True,
            )
            return spec()

    def _uri_to_file_scope(self, uri: Optional[Uri]) -> str:
        folder_uris = list(self._file_config.keys())
        return _uri_to_scope(folder_uris, uri)

    def _uri_to_workspace_scope(self, uri: Optional[Uri]) -> str:
        folder_uris = [f.uri for f in self.workspace.folders.values()]

        if (root_uri := self.workspace.root_uri) is not None:
            folder_uris.append(root_uri)

        return _uri_to_scope(folder_uris, uri)

    def _discover_config_files(self) -> List[pathlib.Path]:
        """Scan the workspace for available configuration files."""
        folder_uris = {f.uri for f in self.workspace.folders.values()}

        if (root_uri := self.workspace.root_uri) is not None:
            folder_uris.add(root_uri)

        paths = []
        for uri in folder_uris:
            if (folder_path := Uri.parse(uri).fs_path) is None:
                continue

            self.logger.debug("Scanning workspace folder: '%s'", folder_path)
            for p in pathlib.Path(folder_path).glob("**/pyproject.toml"):
                self.logger.debug("Found '%s'", p)
                paths.append(p)

        return paths

    def update_file_configuration(self, paths: Optional[List[pathlib.Path]] = None):
        """Update the internal cache of configuration coming from files.

        Parameters
        ----------
        paths
           A list of filepaths to read from.
           If not set, this method will scan each workspace folder for relevant files.
        """
        if paths is None:
            paths = self._discover_config_files()

        for path in paths:
            try:
                data = toml.loads(path.read_text())
                config = dict(esbonio=data.get("tool", {}).get("esbonio", {}))
                scope = str(Uri.for_file(path.parent))

                self._file_config[scope] = config
                self.logger.debug(
                    "File '%s' configuration: %s", scope, json.dumps(config, indent=2)
                )
            except Exception:
                self.logger.error(
                    "Unable to read configuration file: '%s'", exc_info=True
                )

        self._notify_subscriptions()

    async def update_workspace_configuration(self):
        """Update the internal cache of the client's workspace configuration."""
        if not self.supports_workspace_config:
            return

        # Request configuration at the global scope, and at each workspace.
        scopes = [None] + [f.uri for f in self.workspace.folders.values()]
        if (root_uri := self.workspace.root_uri) is not None and root_uri not in scopes:
            scopes.append(root_uri)

        params = types.ConfigurationParams(
            items=[
                types.ConfigurationItem(section="esbonio", scope_uri=scope)
                for scope in scopes
            ]
        )
        self.logger.debug(
            "workspace/configuration: %s",
            json.dumps(self.converter.unstructure(params), indent=2),
        )

        try:
            results = await self.server.get_configuration_async(params)
        except Exception:
            self.logger.error("Unable to get workspace configuration", exc_info=True)
            return

        for scope, result in zip(scopes, results):
            self.logger.debug(
                "Workspace '%s' configuration: %s", scope, json.dumps(result, indent=2)
            )
            # result can be `None`
            result = result or {}

            if "esbonio" not in result:
                result = dict(esbonio=result)

            self._workspace_config[scope or ""] = result

        self._notify_subscriptions()


def _uri_to_scope(known_scopes: List[str], uri: Optional[Uri]) -> str:
    """Convert the given uri to a scope or the empty string if none could be found.

    Parameters
    ----------
    known_scopes
       The collection of known scopes.

    uri
       The uri to convert

    Returns
    -------
    str
        The scope to use
    """
    if uri is None:
        return ""

    uri = uri.resolve()
    known_uris = [str(Uri.parse(u).resolve()) for u in known_scopes]

    candidates = [scope for scope in known_uris if str(uri).startswith(scope)]
    if len(candidates) == 0:
        return ""

    # Return the most specific match.
    return sorted(candidates, key=len, reverse=True)[0]


def _merge_configs(*configs: Dict[str, Any]):
    """Recursively merge all the given configuration sources together.

    The last config given takes precedence.
    """
    final = {}
    all_keys: Set[str] = set()

    for c in configs:
        all_keys.update(c.keys())

    for k in all_keys:
        values = [
            v for c in configs if (v := c.get(k, attrs.NOTHING)) is not attrs.NOTHING
        ]

        # Do we need to recurse?
        if isinstance(values[-1], dict):
            # Be sure to only pass dictionary values to _merge_configs.
            final[k] = _merge_configs(*[v for v in values if isinstance(v, dict)])
        else:
            final[k] = values[-1]

    return final



from __future__ import annotations

import pathlib
import typing

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

if typing.TYPE_CHECKING:
    from typing import Optional
    from typing import Set


EXPECTED = {
    "ref",
    "doc",
    "option",
    "func",
    "class",
    "c:macro",
    "c:func",
    "py:func",
    "py:class",
    "std:ref",
    "std:doc",
}

UNEXPECTED = {
    "macro",
    "restructuredtext-unimplemented-role",
}


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("::", None, None),
        (":", EXPECTED, UNEXPECTED),
        (":r", EXPECTED, UNEXPECTED),
        (":c:func", EXPECTED, UNEXPECTED),
        (":c:func: ", None, None),
        ("  ::", None, None),
        ("  :", EXPECTED, UNEXPECTED),
        ("  :r", EXPECTED, UNEXPECTED),
        ("  :c:func", EXPECTED, UNEXPECTED),
        ("  :c:func: ", None, None),
        ("(:", EXPECTED, UNEXPECTED),
        ("(:r", EXPECTED, UNEXPECTED),
        ("(:c:func", EXPECTED, UNEXPECTED),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_rst_role_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer role completions in rst documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "roles.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        (":ref:`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        (":std:ref:`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        (":doc:`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        (":std:doc:`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        (
            ":class:`",
            {"counters.pattern.PatternCounter", "counters.pattern.NoMatchesError"},
            set(),
        ),
        (
            ":py:class:`",
            {"counters.pattern.PatternCounter", "counters.pattern.NoMatchesError"},
            set(),
        ),
        (":func:`", {"counters.pattern.count_numbers"}, set()),
        (":py:func:`", {"counters.pattern.count_numbers"}, set()),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_rst_role_target_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer role target completions in rst
    documents."""
    test_uri = uri_for("workspaces", "demo", "rst", "roles.rst")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index(".. Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="restructuredtext",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '.. Add your note here...' comment in
    # 'demo/rst/directives.rst' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("{", EXPECTED, UNEXPECTED),
        ("{r", EXPECTED, UNEXPECTED),
        ("{c:func", EXPECTED, UNEXPECTED),
        ("{c:func} ", None, None),
        ("  {", EXPECTED, UNEXPECTED),
        ("  {r", EXPECTED, UNEXPECTED),
        ("  {c:func", EXPECTED, UNEXPECTED),
        ("  {c:func} ", None, None),
        ("({", EXPECTED, UNEXPECTED),
        ("({r", EXPECTED, UNEXPECTED),
        ("({c:func", EXPECTED, UNEXPECTED),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_myst_role_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "roles.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected


@pytest.mark.parametrize(
    "text, expected, unexpected",
    [
        ("{ref}`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        ("{std:ref}`", {"genindex", "modindex", "rst-roles-completion"}, set()),
        ("{doc}`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        ("{std:doc}`", {"demo_myst", "demo_rst", "rst/domains/python"}, set()),
        (
            "{class}`",
            {"counters.pattern.PatternCounter", "counters.pattern.NoMatchesError"},
            set(),
        ),
        (
            "{py:class}`",
            {"counters.pattern.PatternCounter", "counters.pattern.NoMatchesError"},
            set(),
        ),
        ("{func}`", {"counters.pattern.count_numbers"}, set()),
        ("{py:func}`", {"counters.pattern.count_numbers"}, set()),
    ],
)
@pytest.mark.asyncio(scope="session")
async def test_myst_role_target_completions(
    client: LanguageClient,
    uri_for,
    text: str,
    expected: Optional[Set[str]],
    unexpected: Optional[Set[str]],
):
    """Ensure that the language server can offer completions in MyST documents."""
    test_uri = uri_for("workspaces", "demo", "myst", "roles.md")

    uri = str(test_uri)
    fpath = pathlib.Path(test_uri)
    contents = fpath.read_text()
    linum = contents.splitlines().index("% Add your reference here...")

    # Open the file
    client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=uri,
                language_id="markdown",
                version=1,
                text=contents,
            )
        )
    )

    # Write some text
    #
    # This should replace the '% Add your note here...' comment in
    # 'demo/myst/directives.md' with the provided text
    client.text_document_did_change(
        types.DidChangeTextDocumentParams(
            text_document=types.VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[
                types.TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=types.Range(
                        start=types.Position(line=linum, character=0),
                        end=types.Position(line=linum + 1, character=0),
                    ),
                )
            ],
        )
    )

    # Make the completion request
    results = await client.text_document_completion_async(
        types.CompletionParams(
            text_document=types.TextDocumentIdentifier(uri=uri),
            position=types.Position(line=linum, character=len(text)),
        )
    )

    # Close the document - without saving!
    client.text_document_did_close(
        types.DidCloseTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=uri)
        )
    )

    if expected is None:
        assert results is None
    else:
        items = {item.label for item in results.items}
        unexpected = unexpected or set()

        assert expected == items & expected
        assert set() == items & unexpected



from __future__ import annotations

import typing

import pytest

from esbonio.server.features.project_manager import Project

if typing.TYPE_CHECKING:
    from sphinx.application import Sphinx


@pytest.mark.asyncio
async def test_python_domain_discovery(app: Sphinx, project: Project):
    """Ensure that we can correctly index all the objects associated with the Python domain."""

    expected = set()
    domain = app.env.domains["py"]

    for name, _, objtype, docname, _, _ in domain.get_objects():
        expected.add((name, objtype, docname))

    actual = set()
    db = await project.get_db()
    cursor = await db.execute(
        "SELECT name, objtype, docname FROM objects where domain = 'py'"
    )

    for item in await cursor.fetchall():
        actual.add(item)

    assert expected == actual


@pytest.mark.asyncio
async def test_std_domain_discovery(app: Sphinx, project: Project):
    """Ensure that we can correctly index all the objects associated with the Python domain."""

    expected = set()
    domain = app.env.domains["std"]

    for name, _, objtype, docname, _, _ in domain.get_objects():
        expected.add((name, objtype, docname))

    actual = set()
    db = await project.get_db()
    cursor = await db.execute(
        "SELECT name, objtype, docname FROM objects where domain = 'std'"
    )

    for item in await cursor.fetchall():
        actual.add(item)

    assert expected == actual



from __future__ import annotations

import inspect
import logging
import os
import pathlib
import sys
import typing

from . import types
from .types import Uri
from .util import logger

if typing.TYPE_CHECKING:
    from types import ModuleType
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Set
    from typing import Tuple
    from typing import Union


DIAGNOSTIC_SEVERITY = {
    logging.ERROR: types.DiagnosticSeverity.Error,
    logging.INFO: types.DiagnosticSeverity.Information,
    logging.WARNING: types.DiagnosticSeverity.Warning,
}


class DiagnosticFilter(logging.Filter):
    """A logging handler that can extract errors from Sphinx's build output."""

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app = app
        self.diagnostics: Dict[Uri, Set[types.Diagnostic]] = {}

    def get_location(self, location: str) -> Tuple[str, Optional[int]]:
        if not location:
            conf = pathlib.Path(self.app.confdir, "conf.py")
            return (str(conf), None)

        lineno = None
        path, parts = self.get_location_path(location)

        if len(parts) == 1:
            try:
                lineno = int(parts[0])
            except ValueError:
                pass

        if len(parts) == 2 and parts[0].startswith("docstring of "):
            target = parts[0].replace("docstring of ", "")
            lineno = self.get_docstring_location(target, parts[1])

        return (path, lineno)

    def get_location_path(self, location: str) -> Tuple[str, List[str]]:
        """Determine the filepath from the given location."""

        if location.startswith("internal padding before "):
            location = location.replace("internal padding before ", "")

        if location.startswith("internal padding after "):
            location = location.replace("internal padding after ", "")

        path, *parts = location.split(":")

        # On windows the rest of the path will be the first element of parts
        if pathlib.Path(location).drive:
            path += f":{parts.pop(0)}"

        # Diagnostics in .. included:: files are reported relative to the process'
        # working directory, so ensure the path is absolute.
        path = os.path.abspath(path)

        return path, parts

    def get_docstring_location(self, target: str, offset: str) -> Optional[int]:
        # The containing module will be the longest substring we can find in target
        candidates = [m for m in sys.modules.keys() if target.startswith(m)] + [""]
        module = sys.modules.get(sorted(candidates, key=len, reverse=True)[0], None)

        if module is None:
            return None

        obj: Union[ModuleType, Any, None] = module
        dotted_name = target.replace(module.__name__ + ".", "")

        for name in dotted_name.split("."):
            obj = getattr(obj, name, None)
            if obj is None:
                return None

        try:
            _, line = inspect.getsourcelines(obj)  # type: ignore

            # Correct off by one error for docstrings that don't start with a newline.
            nl = (obj.__doc__ or "").startswith("\n")
            return line + int(offset) - (not nl)
        except Exception:
            logger.debug("Unable to determine diagnostic location\n%s", exc_info=True)
            return None

    def filter(self, record: logging.LogRecord) -> bool:
        conditions = [
            "sphinx" not in record.name,
            record.levelno not in {logging.WARNING, logging.ERROR},
        ]

        if any(conditions):
            return True

        loc = getattr(record, "location", "")
        doc, lineno = self.get_location(loc)
        line = lineno or 1

        try:
            # Not every message contains a string...
            if not isinstance(record.msg, str):
                message = str(record.msg)
            else:
                message = record.msg

            # Only attempt to format args if there are args to format
            if record.args is not None and len(record.args) > 0:
                message = message % record.args

        except Exception:
            message = str(record.msg)
            logger.error("Unable to format diagnostic message: %s", exc_info=True)

        diagnostic = types.Diagnostic(
            range=types.Range(
                start=types.Position(line=line - 1, character=0),
                end=types.Position(line=line, character=0),
            ),
            message=message,
            severity=DIAGNOSTIC_SEVERITY.get(
                record.levelno, types.DiagnosticSeverity.Warning
            ),
        )

        self.diagnostics.setdefault(Uri.for_file(doc), set()).add(diagnostic)
        return True



from __future__ import annotations

import logging
import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.server.features import roles
from esbonio.server.features.project_manager import ProjectManager
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    from typing import List
    from typing import Optional

    from esbonio.server import Uri
    from esbonio.server.features.project_manager import Project


TARGET_KINDS = {
    "attribute": lsp.CompletionItemKind.Field,
    "doc": lsp.CompletionItemKind.File,
    "class": lsp.CompletionItemKind.Class,
    "envvar": lsp.CompletionItemKind.Variable,
    "function": lsp.CompletionItemKind.Function,
    "method": lsp.CompletionItemKind.Method,
    "module": lsp.CompletionItemKind.Module,
    "term": lsp.CompletionItemKind.Text,
}


class ObjectsProvider(roles.RoleTargetProvider):
    """Expose domain objects as potential role targets"""

    def __init__(self, logger: logging.Logger, manager: ProjectManager):
        self.manager = manager
        self.logger = logger

    async def suggest_targets(  # type: ignore[override]
        self,
        context: server.CompletionContext,
        *,
        obj_types: List[str],
    ) -> Optional[List[lsp.CompletionItem]]:
        self.logger.debug("Suggesting targets for types: %s", obj_types)

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        db = await project.get_db()
        query = " ".join(
            [
                "SELECT name, display, objtype FROM objects",
                "WHERE printf('%s:%s', domain, objtype) IN (",
                ", ".join("?" for _ in range(len(obj_types))),
                ");",
            ]
        )

        items = []
        cursor = await db.execute(query, tuple(obj_types))
        for name, display, type_ in await cursor.fetchall():
            kind = TARGET_KINDS.get(type_, lsp.CompletionItemKind.Reference)
            items.append(
                lsp.CompletionItem(
                    label=name,
                    detail=display,
                    kind=kind,
                ),
            )

        return items


class SphinxRoles(roles.RoleProvider):
    """Support for roles in a sphinx project."""

    def __init__(self, manager: ProjectManager):
        self.manager = manager

    async def get_default_domain(self, project: Project, uri: Uri) -> str:
        """Get the name of the default domain for the given document"""

        # Does the document have a default domain set?
        results = await project.find_symbols(
            uri=str(uri.resolve()),
            kind=lsp.SymbolKind.Class.value,
            detail="default-domain",
        )
        if len(results) > 0:
            default_domain = results[0][1]
        else:
            default_domain = None

        primary_domain = await project.get_config_value("primary_domain")
        return default_domain or primary_domain or "py"

    async def get_role(self, uri: Uri, name: str) -> Optional[types.Role]:
        """Return the role with the given name."""

        if (project := self.manager.get_project(uri)) is None:
            return None

        if (role := await project.get_role(name)) is not None:
            return role

        if (role := await project.get_role(f"std:{name}")) is not None:
            return role

        default_domain = await self.get_default_domain(project, uri)
        return await project.get_role(f"{default_domain}:{name}")

    async def suggest_roles(
        self, context: server.CompletionContext
    ) -> Optional[List[types.Role]]:
        """Given a completion context, suggest roles that may be used."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        default_domain = await self.get_default_domain(project, context.uri)

        result: List[types.Role] = []
        for name, implementation in await project.get_roles():
            # std: directives can be used unqualified
            if name.startswith("std:"):
                short_name = name.replace("std:", "")
                result.append(
                    types.Role(name=short_name, implementation=implementation)
                )

            # Also suggest unqualified versions of directives from the currently active domain.
            elif name.startswith(f"{default_domain}:"):
                short_name = name.replace(f"{default_domain}:", "")
                result.append(
                    types.Role(name=short_name, implementation=implementation)
                )

            result.append(types.Role(name=name, implementation=implementation))

        return result


def esbonio_setup(
    esbonio: server.EsbonioLanguageServer,
    project_manager: ProjectManager,
    roles_feature: roles.RolesFeature,
):
    role_provider = SphinxRoles(project_manager)
    obj_provider = ObjectsProvider(
        esbonio.logger.getChild("ObjectsProvider"), project_manager
    )

    roles_feature.add_role_provider(role_provider)
    roles_feature.add_target_provider("objects", obj_provider)



from __future__ import annotations

import inspect
import typing

import attrs

from esbonio import server

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Coroutine
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Union


@attrs.define
class Directive:
    """Represents a directive."""

    name: str
    """The name of the directive, as the user would type in an rst file."""

    implementation: Optional[str]
    """The dotted name of the directive's implementation."""


class DirectiveProvider:
    """Base class for directive providers"""

    def suggest_directives(
        self, context: server.CompletionContext
    ) -> Union[
        Optional[List[Directive]], Coroutine[Any, Any, Optional[List[Directive]]]
    ]:
        """Given a completion context, suggest directives that may be used."""
        return None


class DirectiveFeature(server.LanguageFeature):
    """'Backend' support for directives.

    It's this language feature's responsibility to provide an API that exposes the
    information a "frontend" language feature may want.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._providers: Dict[int, DirectiveProvider] = {}

    def add_provider(self, provider: DirectiveProvider):
        """Register a directive provider.

        Parameters
        ----------
        provider
           The directive provider
        """
        self._providers[id(provider)] = provider

    async def suggest_directives(
        self, context: server.CompletionContext
    ) -> List[Directive]:
        """Suggest directives that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context.
        """
        items: List[Directive] = []

        for provider in self._providers.values():
            try:
                result: Optional[List[Directive]] = None

                aresult = provider.suggest_directives(context)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result:
                    items.extend(result)
            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.suggest_directives'", name, exc_info=True
                )

        return items


def esbonio_setup(server: server.EsbonioLanguageServer):
    directives = DirectiveFeature(server)
    server.add_feature(directives)
